__author__ = "DCollins"

import sys

# sys.path.append("../AcraNetwork")
import os

import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.iNetX as inetx
import AcraNetwork.MPEGTS as MPEGTS
import base64
import AcraNetwork.MPEG.PMT as MPEGPMT
import AcraNetwork.MPEG.PES as pes
import struct
import datetime

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


mpeg_ts_repr = """PID=0X0 PUSI=True TSC=Not Scrambled Adaption=Payload Only
PID=0X1000 PUSI=True TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=True TSC=Not Scrambled Adaption=Adaption and Payload
 Adaption=Discontunity=False, random=True Elementary Stream Indicator=False PCR=True OPCR=False Splicing Point Flag=False Transport Private Data=False, Adaption Extension=False
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
"""


class MPEGTSBasicTest(unittest.TestCase):

    ######################
    # Read a complete pcap file
    ######################

    def test_readFirstMPEGTS(self):
        """
        Very simple test that reads a pcap file with mpegts packets.
        Takes the first packet in there and decoms the mpegts blocks
        Verifies each block in that first packet
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        mypcaprecord = p[0]

        ethpacket = SimpleEthernet.Ethernet()  # Create an Ethernet object
        ethpacket.unpack(mypcaprecord.packet)  # Unpack the pcap record into the eth object
        ippacket = SimpleEthernet.IP()  # Create an IP packet
        ippacket.unpack(ethpacket.payload)  # Unpack the ethernet _payload into the IP packet
        udppacket = SimpleEthernet.UDP()  # Create a UDP packet
        udppacket.unpack(ippacket.payload)  # Unpack the IP _payload into the UDP packet
        inetxpacket = inetx.iNetX()  # Create an iNetx object
        inetxpacket.unpack(udppacket.payload)  # Unpack the UDP _payload into this iNetX object

        mpegts = MPEGTS.MPEGTS()
        mpegts.unpack(inetxpacket.payload)
        self.assertEqual(len(mpegts), 7)
        for packet_index in list(range(7)):
            if packet_index == 0:
                self.assertEqual(mpegts.blocks[packet_index].pid, 0)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
            elif packet_index == 1:
                self.assertEqual(mpegts.blocks[packet_index].pid, 4096)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
            else:
                self.assertEqual(mpegts.blocks[packet_index].pid, 256)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
        p.close()

    def test_pack_unpack_adaption(self):
        ad = MPEGTS.MPEGAdaption()
        adex = MPEGTS.MPEGAdaptionExtension()
        ad.adaption_extension = adex
        packed = ad.pack()
        ad2 = MPEGTS.MPEGAdaption()
        ad2.unpack(packed)
        self.assertEqual(ad2, ad)

    def test_pack_unpack(self):
        """
        Very simple test that reads a pcap file with mpegts packets.
        Takes the first packet in there and decoms the mpegts blocks
        Verifies each block in that first packet
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        mypcaprecord = p[0]
        self.maxDiff = None
        mpegts = MPEGTS.MPEGTS()
        payload = mypcaprecord.payload[0x46:]
        mpegts.unpack(payload)
        self.assertEqual(mpeg_ts_repr, repr(mpegts))
        self.assertEqual(len(mpegts), 7)
        p.close()
        for block in mpegts:
            if block.adaption_field is not None:
                pass
        packed = mpegts.pack()
        self.assertEqual(packed, payload)

        mpegts2 = MPEGTS.MPEGTS()
        mpegts2.unpack(packed)
        self.assertEqual(mpegts, mpegts2)

    def test_readAllMPEGTS(self):
        """
        Reads the same mpeg ts file as previously but reads all the data
        in the file and checks for any continuity errors
        :return:
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        for mypcaprecord in p:

            ethpacket = SimpleEthernet.Ethernet()  # Create an Ethernet object
            ethpacket.unpack(mypcaprecord.packet)  # Unpack the pcap record into the eth object
            ippacket = SimpleEthernet.IP()  # Create an IP packet
            ippacket.unpack(ethpacket.payload)  # Unpack the ethernet _payload into the IP packet
            udppacket = SimpleEthernet.UDP()  # Create a UDP packet
            udppacket.unpack(ippacket.payload)  # Unpack the IP _payload into the UDP packet
            inetxpacket = inetx.iNetX()  # Create an iNetx object
            inetxpacket.unpack(udppacket.payload)  # Unpack the UDP _payload into this iNetX object
            mpegts = MPEGTS.MPEGTS()
            mpegts.unpack(inetxpacket.payload)
            self.assertEqual(len(mpegts), 7)

        p.close()

    @unittest.skip("Broken")
    def test_stanag(self):
        ts_file = open(os.path.join(THIS_DIR, "stanag_sample.ts"), mode="rb")
        h264_data = MPEGTS.H264()
        self.assertTrue(h264_data.unpack(ts_file.read()))
        ts_file.close()
        self.assertEqual(len(h264_data.nals), 258)
        nal_counts = {}
        for nal in h264_data.nals:
            if not nal.type in nal_counts:
                nal_counts[nal.type] = 1
            else:
                nal_counts[nal.type] += 1
        self.assertEqual(nal_counts[0], 35)
        self.assertEqual(nal_counts[6], 70)


# Take from Wireshar - > cop
pmt_payload = base64.b64decode(
    "R0EAGQACsDgAAcEAAOEB8AAb4QLwAAPhA/AAFeEE8BwFBEtMVkEmCQEA/0tMVkEADycJwQAAwIAAwQAAX+WEI/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////8="
)


class MPEGPNT(unittest.TestCase):
    def test_pmt_unpack(self):
        pmt = MPEGPMT.MPEGPacketPMT()
        pmt.unpack(pmt_payload)

        pmt2 = MPEGPMT.MPEGPacketPMT()
        pmt2.unpack(pmt.pack())
        self.assertEqual(pmt, pmt2)
        self.assertEqual(pmt_payload, pmt2.pack())

    def test_pmt_random(self):
        pmt = MPEGPMT.MPEGPacketPMT()
        pmt.sync = 0x47
        pmt.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_ONLY
        pmt.pid = 0x100
        pmt.continuitycounter = 7
        pmt.last_section = 0xE
        pmt.section = 0xD
        pmt.program_number = 0xDEAD
        t = MPEGPMT.DescriptorTag()
        t.data = struct.pack(">5H", *(list(range(5))))
        t.tag = 0xA
        s = MPEGPMT.PMTStream()
        s.descriptor_tags = [t]
        s.elementary_pid = 0x1FEF
        pmt.streams = [s]
        _packed = pmt.pack()

        pmt2 = MPEGPMT.MPEGPacketPMT()
        self.assertTrue(pmt2.unpack(_packed))
        self.assertEqual(pmt, pmt2)


pes_packet = base64.b64decode(
    "R0EEP4UA////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////AAAB/AAsgYAFIQQD/tEAD98AHwYOKzQCCwEBDgEDAQEAAAAOAggABg/Gi5FmYwECH1M="
)


class MPEG_PES(unittest.TestCase):
    def test_pes_unpack(self):
        self.assertEqual(188, len(pes_packet))  # Verify the length
        p = pes.PES()
        p.unpack(pes_packet)  # unpack as PES
        self.assertEqual(50, len(p.payload))
        # print(repr(p))

        self.assertEqual(36, len(p.pesdata))  # Verify the pes payload
        self.assertEqual(pes_packet, p.pack())  # Verify that pack turns it back exactly into the same packet
        p2 = pes.PES()
        p2.unpack(p.pack())  # Unpack it again
        self.assertEqual(p, p2)  # Compre packets
        self.assertEqual(p2.pack(), pes_packet)  # Verify packing

    def test_stanag_unpack(self):
        p = pes.STANAG4609()
        p.unpack(pes_packet)
        print(repr(p))
        self.assertEqual(36, len(p.pesdata))
        self.assertEqual(len(p.pack()), 188)
        self.assertEqual(p.pack(), pes_packet)
        print(repr(p.time))

    def test_stanag_create(self):
        ref = pes.STANAG4609()
        ref.unpack(pes_packet)
        # Build a packet from scratch that matches the refernce packet
        p = pes.STANAG4609()
        p.pid = pes.STANAG4609_PID
        p.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_AND_ADAPTION
        p.adaption_field = MPEGTS.MPEGAdaption()
        p.adaption_field.length = 133
        p.time = datetime.datetime(2024, 1, 25, 15, 7, 59, 767139, tzinfo=datetime.timezone.utc)
        p.header_data = pes.ts_to_buf(187.14)
        p.extension_w1 = pes.PES_EXTENSION_W1
        p.extension_w2 = pes.PES_EXTENSION_W2
        p.streamid = 0xFC
        p.stanag_counter = 15
        p.continuitycounter = 15
        p.pack()
        self.assertEqual(p, ref)
        self.assertEqual(p.pack(), pes_packet)

    def test_stanag_pts(self):
        ts = 0x210403FED1
        exp = 187.14
        self.assertEqual(pes.pts_to_ts(ts), exp)
        _conv = pes.ts_to_pts(exp)
        print(f"{_conv:#0X}")
        self.assertEqual(_conv, ts)


corrupt_vid_ = base64.b64decode(
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEdAEREABBYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsdeSR2TR0AQEQAFyMnKy8zNzs/Q0dLT1NXW19jZ2tvc3d7f4OHi4+Tl5ufo6err7O3u7/Dx8vP09fb3+Pn6+/z9/v8AAQIDBAUGBwgJCgsMDQ4PEBESExQVFhcYGRobHB0eHyAhIiMkJSYnKCkqKywtLi8wMTIzNDU2Nzg5Ojs8PT4/QEFCQ0RFRkdISUpLTE1OT1BRUlNUVVZXWFlaW1xdXl9gYWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eRnP0dRHH/8QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEcf/xAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAARx//EAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABHH/8QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEcf/xAAAAAAAAAAAAAAAAAAAAAA"
)


class CorruptVid(unittest.TestCase):
    pass


if "unittest.util" in __import__("sys").modules:
    # Show full diff in self.assertEqual.
    __import__("sys").modules["unittest.util"]._MAX_LENGTH = 999999999

if __name__ == "__main__":
    unittest.main()



import struct
from functools import reduce
import logging
from datetime import datetime, timezone
from enum import IntEnum
from decimal import Decimal


DATA_TYPE_TIMEFMT_1 = 0x11
DATA_TYPE_TIMEFMT_2 = 0x12
DATA_TYPE_PCM_DATA_FMT1 = 0x9
DATA_TYPE_MILSTD1553_FMT1 = 0x19
DATA_TYPE_ARINC429_FMT0 = 0x38
DATA_TYPE_UART_FMT0 = 0x50
DATA_TYPE_COMPUTER_GENERATED_FORMAT_0 = 0x0
DATA_TYPE_COMPUTER_GENERATED_FORMAT_1 = 0x1
DATA_TYPE_ANALOG = 0x21
DATA_TYPE_VIDEO_FMT0 = 0x40
DATA_TYPE_VIDEO_FMT1 = 0x41
DATA_TYPE_VIDEO_FMT2 = 0x42
DATA_TYPE_VIDEO_FMT3 = 0x43


class DataType(IntEnum):
    TIMEFORMAT_1 = DATA_TYPE_TIMEFMT_1
    TIMEFORMAT_2 = DATA_TYPE_TIMEFMT_2
    PCM = DATA_TYPE_PCM_DATA_FMT1
    MILSTD1553 = DATA_TYPE_MILSTD1553_FMT1
    ARINC429 = DATA_TYPE_ARINC429_FMT0
    COMPUTER_FORMAT_0 = DATA_TYPE_COMPUTER_GENERATED_FORMAT_0
    COMPUTER_FORMAT_1 = DATA_TYPE_COMPUTER_GENERATED_FORMAT_1
    ANALOG = DATA_TYPE_ANALOG
    UART = DATA_TYPE_UART_FMT0
    VIDEO_FORMAT_0 = DATA_TYPE_VIDEO_FMT0
    VIDEO_FORMAT_1 = DATA_TYPE_VIDEO_FMT1
    VIDEO_FORMAT_2 = DATA_TYPE_VIDEO_FMT2
    VIDEO_FORMAT_3 = DATA_TYPE_VIDEO_FMT3


TS_RTC = 0
TS_SECONDARY = 1

TS_CH4 = 0
TS_IEEE1558 = 1
TS_ERTC = 2


def buf_to_printable(buffer):
    pw = ""
    for idx in range(len(buffer)):
        (v,) = struct.unpack_from(">B", buffer, idx)
        pw += "{:#04X} ".format(v)
        if idx % 8 == 7:
            pw += "\n"
    return pw


class PTPTime(object):
    def __init__(self, seconds=0, nanoseconds=0):
        self.seconds = seconds
        self.nanoseconds = nanoseconds

    def pack(self):
        return struct.pack("<II", self.nanoseconds, self.seconds)

    def unpack(self, buffer):
        (self.nanoseconds, self.seconds) = struct.unpack("<II", buffer)
        return True

    def to_rtc(self):
        ptp_as_date = datetime.fromtimestamp(self.seconds, tz=timezone.utc)
        start_of_year = datetime(ptp_as_date.year, 1, 1, 0, 0, 0)
        start_of_year = start_of_year.replace(tzinfo=timezone.utc)
        seconds_since_start_year = int((ptp_as_date - start_of_year).total_seconds())
        rtc_time = int(seconds_since_start_year * 1e7 + self.nanoseconds / 100)
        return rtc_time

    def to_pinksheet_rtc(self):
        ptp_vector = (Decimal(self.seconds) * Decimal(1e9)) + Decimal(self.nanoseconds)
        ptp_vector_100ns = ptp_vector // Decimal(100)
        ptp_vector_truncated = int(ptp_vector_100ns) & (pow(2, 48) - 1)
        return ptp_vector_truncated

    def __repr__(self):
        time_fmt = "%H:%M:%S %d-%b %Y"
        ts = datetime.fromtimestamp(self.seconds, tz=timezone.utc)
        date_str = ts.strftime(time_fmt)
        return "PTP: {} nanosec={}".format(date_str, self.nanoseconds)

    def __eq__(self, __value):
        if not isinstance(__value, PTPTime):
            return False
        if self.nanoseconds != __value.nanoseconds or self.seconds != __value.seconds:
            return False
        return True

    def __add__(self, val):
        if isinstance(val, PTPTime):
            addns = self.nanoseconds + val.nanoseconds
            ns = int(addns % 1e9)
            sec = self.seconds + val.seconds + int(addns // 1e9)
            return PTPTime(sec, ns)
        else:
            raise Exception("Addition of other PTPTime instances")

    def __sub__(self, val):
        if isinstance(val, PTPTime):
            if val.nanoseconds > self.nanoseconds:
                sec = self.seconds - val.seconds - 1
                nsec = int(1e9) - (val.nanoseconds - self.nanoseconds)
            else:
                sec = self.seconds - val.seconds
                nsec = self.nanoseconds - val.nanoseconds
            return PTPTime(sec, nsec)
        else:
            raise Exception("Subtraction of other PTPTime instances")

    def __lt__(self, val):
        if isinstance(val, PTPTime):
            return (self.seconds + self.nanoseconds / 1e9) < (val.seconds + val.nanoseconds / 1e9)

    def __le__(self, val):
        if isinstance(val, PTPTime):
            _d = (self.seconds + self.nanoseconds / 1e9) <= (val.seconds + val.nanoseconds / 1e9)
            return _d


class RTCTime(object):
    def __init__(self, count=0):
        self.count = count

    def pack(self):
        msw = (self.count >> 32) & 0xFFFF
        lsw = self.count & 0xFFFFFFFF
        return struct.pack("<IHH", lsw, msw, 0)

    def unpack(self, buffer):
        (lsw, msw, _zero) = struct.unpack("<IHH", buffer)
        self.count = lsw + (msw << 32)
        return True

    def to_rtc(self):
        return self.count

    def to_pinksheet_rtc(self):
        return self.count

    def __repr__(self):
        return "RTC: count={}".format(self.count)

    def __eq__(self, __value):
        if not isinstance(__value, RTCTime):
            return False
        if self.count != __value.count:
            return False
        return True



"""
.. module:: Chapter10
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct Chapter10 Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
from functools import reduce
from . import TS_CH4, TS_ERTC, TS_IEEE1558, TS_RTC, TS_SECONDARY, RTCTime, PTPTime
import logging
import typing


logger = logging.getLogger(__name__)


def get_checksum_buf(buf):
    """
    Return the arithmetic checksum of a header

    :param buf:
    :return:
    """
    if len(buf) % 2 != 0:
        raise Exception("buffer needs to be 16-bit aligned")

    words = struct.unpack("<{}H".format(len(buf) // 2), buf)
    sum = reduce(lambda x, y: x + y, words)

    return sum % 65536


def get_checksum_byte_buf(buf):
    """
    Return the arithmetic checksum of a header

    :param buf:
    :return:
    """

    checksumbytes = struct.unpack(f"<{len(buf)}B", buf)
    sum = reduce(lambda x, y: x + y, checksumbytes)

    return sum % 65536


class Chapter10(object):
    """
    Class to pack and unpack Chapter10 payloads.

    Create a packet and transmit it via UDP

    >>> import socket
    >>> # Open a socket
    >>> tx_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> # Create the Ch10 UDP wrapper
    >>> ch10_udp = Chapter10UDP()
    >>> ch10_udp.type = Chapter10UDP.TYPE_FULL
    >>> ch10_udp.sequence = 1
    >>> # Populate the Chapter 10 packet inthe wrapper
    >>> ch10_udp.chapter10.channelID = 1
    >>> ch10_udp.chapter10.datatypeversion = 2
    >>> ch10_udp.chapter10.sequence = 3
    >>> ch10_udp.chapter10.packetflag = 0 # No secondary
    >>> ch10_udp.chapter10.datatype = 4
    >>> ch10_udp.chapter10.relativetimecounter = 100
    >>> ch10_udp.chapter10.payload = struct.pack(">II", 33, 44)
    >>> # Send the packet
    >>> tx_socket.sendto(ch10_udp.pack(), ("127.0.0.1", 8010))
    36

    :type syncpattern: int
    :type channelID: int
    :type packetlen: int
    :type datalen: int
    :type datatypeversion: int
    :type sequence: int
    :type datatype: int
    :type relativetimecounter: int
    :type timestamp: int
    :type ts_source: int
    :type payload: bytes
    :type data_checksum_size: int
    :type filler: bytes
    """

    SYNC_WORD = 0xEB25  # :(Object Constant) Sync word

    CH10_HDR_FORMAT = "<HHIIBBBBIHH"
    CH10_HDR_FORMAT_LEN = struct.calcsize(CH10_HDR_FORMAT)

    CH10_OPT_HDR_FORMAT = "<IIHH"
    CH10_OPT_HDR_FORMAT_LEN = struct.calcsize(CH10_OPT_HDR_FORMAT)

    TS_SOURCES = [TS_RTC, TS_SECONDARY]  # :(Object Constant) Valid timesources, assign to :attr:`Chapter10.ts_source`
    TS_SECONDARY_SOURCES = [TS_CH4, TS_IEEE1558, TS_ERTC]

    PKT_FLAG_SECONDARY = 0x80  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_SEC_HDR_TIME = 0x40  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RTC_SYNC_ERROR = 0x20  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_DATA_OVERFLOW = 0x10  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RCC_TIME = 0x0  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_1588_TIME = 0x04  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_8BIT_CHKSUM = 0x1  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_16BIT_CHKSUM = 0x2  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_32BIT_CHKSUM = 0x3  # :(Object Constant) add to :attr:`Chapter10.packetflag` to enable

    def __init__(self):
        self.syncpattern = (
            Chapter10.SYNC_WORD
        )  # :(2 Bytes) contains a static sync value for the every packet. The Packet Sync Pattern value shall be 0xEB25
        self.channelID = 0  # :(2 Bytes) contains a value representing the Packet Channel ID.
        self.packetlen = 0  # :(4 Bytes) contains a value representing the length of the entire packet. The value shall be in bytes and is always a multiple of four
        self.datalen = 0  # :(4 Bytes) contains a value representing the valid data length within the packet
        self.datatypeversion = 0x5  #: RCC released versions
        self.sequence = 0  # :(1 Byte) contains a value representing the packet sequence number for each Channel ID.
        self._packetflag = (
            0  # :(1 Byte) contains bits representing information on the content and format of the packet(s)
        )
        self.datatype = 0  # :(1 Byte) contains a value representing the type and format of the data
        self.relativetimecounter = 0  # :(6 Bytes) contains a value representing the 10 MHz Relative Time Counter (RTC)
        self.ptptime = PTPTime()  #: PTP Timestamp
        self.ts_source = TS_RTC  # :The timestamp source. Select from :attr:`Chapter10.TS_SOURCES`
        self.payload = b""  # :The payload
        self.data_checksum_size = 0
        self.filler = b""
        self.has_secondary_header = False

    @property
    def packetflag(self):
        """
        (1 Byte) contains bits representing information on the content and format of the packet(s)

        :type val: intx
        """
        return self._packetflag

    @packetflag.setter
    def packetflag(self, val):
        if val > 0xFF:
            raise Exception("Packet flag ={:#0X} to valuid".format(val))
        self._packetflag = val

        if self._packetflag >> 7 == 1:
            if ((self._packetflag >> 2) & 0x3) == 0:
                self.ts_source = TS_CH4
            elif ((self._packetflag >> 2) & 0x3) == 1:
                self.ts_source = TS_IEEE1558
            else:
                raise Exception("Time format is illegal")

            self.has_secondary_header = True
        else:
            self.has_secondary_header = False
            self.ts_source = TS_RTC

    def pack(self):
        """
        Pack the Chapter10 object into a binary buffer

        :rtype: bytes
        """

        if self.has_secondary_header:
            if self.ts_source == TS_CH4:
                raise Exception("Ch4 Timestamp in secondary header not supported")

            sec_hdr = self.ptptime.pack() + struct.pack(">HH", 0, 0)
            # Replace the checksum
            cs = get_checksum_byte_buf(sec_hdr)
            sec_hdr = sec_hdr[:-2] + struct.pack("<H", cs)
        else:
            sec_hdr = b""

        total_len_excl_filler = (
            len(sec_hdr) + len(self.payload) + Chapter10.CH10_HDR_FORMAT_LEN + self.data_checksum_size
        )

        # Add the filler
        if total_len_excl_filler % 4 == 0:
            self.filler = b""
        else:
            fill_len = 4 - (total_len_excl_filler % 4)
            fill_val = [0xFF] * fill_len
            self.filler = struct.pack(">{}B".format(fill_len), *fill_val)

        self.packetlen = total_len_excl_filler + len(self.filler)
        self.datalen = len(self.payload)

        checksum = 0
        _rtc_lwr = self.relativetimecounter & 0xFFFFFFFF
        _rtc_upr = self.relativetimecounter >> 32

        hdr = struct.pack(
            Chapter10.CH10_HDR_FORMAT,
            self.syncpattern,
            self.channelID,
            self.packetlen,
            self.datalen,
            self.datatypeversion,
            self.sequence,
            self.packetflag,
            self.datatype,
            _rtc_lwr,
            _rtc_upr,
            checksum,
        )
        hdr = hdr[:-2] + struct.pack("<H", get_checksum_buf(hdr))
        return hdr + sec_hdr + self.payload + self.filler

    def unpack(self, buffer):
        """
        Unpack a string buffer into an Chapter10 object

        :param buffer: A string buffer representing an Chapter10 packet
        :type buffer: bytes
        :rtype: None
        """
        (
            self.syncpattern,
            self.channelID,
            self.packetlen,
            self.datalen,
            self.datatypeversion,
            self.sequence,
            self.packetflag,
            self.datatype,
            _rtc_lwr,
            _rtc_upr,
            checksum,
        ) = struct.unpack_from(Chapter10.CH10_HDR_FORMAT, buffer)

        offset_hdr = struct.calcsize(Chapter10.CH10_HDR_FORMAT) - 2
        exp_checksum = get_checksum_buf(buffer[:offset_hdr])
        if checksum != exp_checksum:
            raise Exception("Ch10 Header checksum {:#0X} does not match expected={:#0X}".format(checksum, exp_checksum))
            # print("Ch10 Header checksum does not match expected={:#0X}".format(exp_checksum))

        self.relativetimecounter = _rtc_lwr + (_rtc_upr << 32)

        if (self.packetflag >> 7) == 1:
            self.has_secondary_header = True
            pkt_hdr_time = (self.packetflag >> 2) & 0x3
            (ts_ns, ts_s, _res, _checksum_sec) = struct.unpack_from(
                Chapter10.CH10_OPT_HDR_FORMAT, buffer, Chapter10.CH10_HDR_FORMAT_LEN
            )

            sec_exp_checksum = get_checksum_byte_buf(
                buffer[
                    Chapter10.CH10_HDR_FORMAT_LEN : Chapter10.CH10_HDR_FORMAT_LEN
                    + Chapter10.CH10_OPT_HDR_FORMAT_LEN
                    - 2
                ]
            )

            if _checksum_sec != sec_exp_checksum:
                raise Exception(
                    "Ch10 Secondary Header checksum ({:#0X}) does not match expected={:#0X}".format(
                        _checksum_sec, sec_exp_checksum
                    )
                )
                # print("Ch10 Secondary Header checksum does not match expected={:#0X}".format(sec_exp_checksum))

            if pkt_hdr_time == 0:
                raise Exception("Ch4 Timestamp in secondary header not supported")
                # print("Ch4 Timestamp in secondary header not supported")
            elif pkt_hdr_time == 1:
                self.ts_source = TS_IEEE1558
                self.ptptime = PTPTime(ts_s, ts_ns)
            else:
                raise Exception("Secondary Header Time Format not legal")
                # print("Secondary Header Time Format not legal")
            self.payload = buffer[(Chapter10.CH10_HDR_FORMAT_LEN + Chapter10.CH10_OPT_HDR_FORMAT_LEN) :]
        else:
            self.has_secondary_header = False
            self.payload = buffer[Chapter10.CH10_HDR_FORMAT_LEN :]
            self.ts_source = TS_RTC

        return True

    def __eq__(self, other):
        if not isinstance(other, Chapter10):
            return False

        _match_att = (
            "syncpattern",
            "channelID",
            "packetlen",
            "datalen",
            "datatypeversion",
            "datatype",
            "_packetflag",
            "relativetimecounter",
            "ptptime",
            "ts_source",
            "payload",
            "data_checksum_size",
            "filler",
            "has_secondary_header",
        )

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "Chapter 10: ChannelID={} Sequence={} DataLen={}".format(self.channelID, self.sequence, self.datalen)


class FileParser(object):
    """
    Parse a Chapter10 file. Open the file and iterate through it
    """

    def __init__(self, filename, mode="rb"):
        self.filename = filename
        self._mode = mode
        self.insync = False
        self._offset = 0
        self._fd = None

    def write(self, ch10packet):
        # type: (Chapter10) -> None
        """
        Write a chapter10 packet to the file
        """
        if not isinstance(ch10packet, Chapter10):
            raise Exception("Only write Chapter10 instances to the file")
        if self._fd is None or self._mode != "wb":
            raise Exception("File name not defined")
        if not self._fd.writable():
            raise Exception("File {} not open for writing".format(self.filename))
        self._fd.write(ch10packet.pack())

    def __enter__(self):
        self._fd = open(self.filename, self._mode)
        return self

    def __exit__(self, type, value, traceback):
        # Exception handling here
        self._fd.close()

    def close(self):
        if self._fd is not None:
            self._fd.close()

    def __iter__(self):
        return self

    def next(self):
        in_sync = False
        pkt_len = 0
        while not in_sync:
            try:
                self._fd.seek(self._offset)
                _first_few_words = self._fd.read(8)
            except:
                raise StopIteration
            try:
                (sync, chid, pkt_len) = struct.unpack("<HHI", _first_few_words)
            except Exception as e:
                logger.debug("Exiting loop err={}".format(e))
                raise StopIteration

            if sync == Chapter10.SYNC_WORD:
                in_sync = True
            else:
                self._offset += 1

        self._fd.seek(self._offset)
        pkt_payload = self._fd.read(pkt_len)

        ch10 = Chapter10()
        try:
            ch10.unpack(pkt_payload)
        except Exception as e:
            logger.error("Failed to unpack data from {}. err={}".format(self._offset, e))
            raise StopIteration
        else:
            self._offset += pkt_len
            return ch10

    __next__ = next



import socket
import logging
import AcraNetwork.iNetX as inetx
import typing
import struct


PCM_HDR_LEN = 10


def string_matching_boyer_moore_horspool(text="", pattern=""):
    """
    Returns positions where pattern is found in text.
    O(n)
    Performance: ord() is slow so we shouldn't use it here
    Example: text = 'ababbababa', pattern = 'aba'
         string_matching_boyer_moore_horspool(text, pattern) returns [0, 5, 7]
    @param text text to search inside
    @param pattern string to search for
    @return list containing offsets (shifts) where pattern is found inside text
    """
    m = len(pattern)
    n = len(text)
    offsets = []
    if m > n:
        return offsets
    skip = []
    for k in range(256):
        skip.append(m)
    for k in range(m - 1):
        my = pattern[k]
        skip[pattern[k]] = m - k - 1

    skip = tuple(skip)
    k = m - 1
    while k < n:
        j = m - 1
        i = k
        while j >= 0 and text[i] == pattern[j]:
            j -= 1
            i -= 1
        if j == -1:
            offsets.append(i + 1)
        k += skip[text[k]]

    return offsets


class SamDec008(object):
    def __init__(self, udp_port: int, timeout: float = 5.0, localaddress=""):

        self.recv_sockets = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.recv_sockets.settimeout(timeout)
        self.recv_sockets.bind((localaddress, udp_port))

        self._current_aligned = True
        self._payload_offset = 0
        self._sequence = None

        self.streamid = 0x153  #: StreamID on which to capture the SAM/DEC data
        self.frame_length = None  #: Calculate the frame length by searching the payload for the sync words
        # For the setup side. Most of the defaults are ok
        self.sync_word = 0xFE6B2840

        # Logging
        self._logger = logging.getLogger(__name__)

    def close(self):
        self.recv_sockets.close()

    def _get_data(self) -> typing.Generator[bytes, None, None]:
        """
        Get the data, agnostic to network vs pcap

        :rtype: collections.Iterable[str]
        """
        while True:
            try:
                data, addr = self.recv_sockets.recvfrom(10000)
            except Exception as e:
                raise Exception(e)
            else:
                yield data

    def frames(self) -> typing.Generator[bytes, None, None]:
        """Get the data from the underlying source, combine the IP fragments and then pull out the payload from the
        inetx packets and align them

        Raises:
            Exception: Cant unpacket the payload

        Yields:
            bytes: the payload captured
        """
        sync_packed = struct.pack(">I", self.sync_word)
        for udp_payload in self._get_data():
            inetx_pkt = inetx.iNetX()
            try:
                inetx_pkt.unpack(udp_payload)
            except Exception as e:
                continue
            else:
                self._payload_offset = PCM_HDR_LEN  # The SAM DEC inserts some header
                if inetx_pkt.streamid == self.streamid:
                    if self._sequence is not None:
                        if self._sequence + 1 % pow(2, 64) == inetx_pkt.sequence:
                            logging.warning(
                                "Missing Sequence number at {}. Is SAM/DEC dropping?".format(inetx_pkt.sequence)
                            )
                    # Start with the previous segment and the current payload
                    payload = inetx_pkt.payload

                    if self.frame_length is None:
                        offset = string_matching_boyer_moore_horspool(payload, sync_packed)
                        if len(offset) == 0:
                            raise Exception("No Frame sync found")
                        elif len(offset) == 1:
                            self.frame_length = len(payload) - PCM_HDR_LEN
                        else:
                            self.frame_length = offset[1] - offset[0]

                    # If we get to within the last 4 bytes then hold it until the next segment
                    while (self._payload_offset + self.frame_length) <= len(payload):
                        # We have enough for a full frame
                        frame_buffer = payload[self._payload_offset : self._payload_offset + self.frame_length]
                        self._payload_offset += self.frame_length
                        # Verify that we are still in sync
                        if frame_buffer[:4] != sync_packed:
                            # Not in sync.
                            print("ERROR:Fell out of alignment at offset {}".format(self._payload_offset))
                            self.frame_length = None
                        else:
                            yield frame_buffer
